{
	Copyright (c) 2020 Adrian Siekierka

	Based on a reconstruction of code from ZZT,
	Copyright 1991 Epic MegaGames, used with permission.

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all
	copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
	SOFTWARE.
}

{$I-}
{$V-}
unit Game;

interface
	uses GameVars, TxtWind;
	const
		PROMPT_NUMERIC = 0;
		PROMPT_ALPHANUM = 1;
		PROMPT_ANY = 2;
	procedure SidebarClearLine(y: integer);
	procedure SidebarClear;
	procedure AdvancePointer(var address: pointer; count: integer);
	procedure BoardClose;
	procedure BoardOpen(boardId: integer);
	procedure BoardChange(boardId: integer);
	procedure BoardCreate;
	procedure WorldCreate;
	procedure TransitionDrawToFill(chr: char; color: integer);
	procedure OopLaserBeam(x, y, dx, dy: integer);
	procedure BoardDrawTile(x, y: integer);
	procedure BoardDrawBorder;
	procedure TransitionDrawToBoard;
	procedure SidebarPromptSlider(editable: boolean; x, y: integer; prompt: string; var value: byte);
	procedure PromptString(x, y, arrowColor, color, width: integer; mode: byte; var buffer: TString50);
	function SidebarPromptYesNo(message: string; defaultReturn: boolean): boolean;
	procedure SidebarPromptString(prompt: string; extension: TString50; var filename: string; promptMode: byte);
	procedure PauseOnError;
	function DisplayIOError: boolean;
	procedure WorldUnload;
	function WorldLoad(filename, extension: TString50): boolean;
	procedure WorldSave(bank: integer);
	procedure WorldSwap(bank, brd: integer);
	function WorldTest(bank: integer): boolean;
	function SaveTest: boolean;
	procedure CopyStatDataToTextWindow(statId: integer; var state: TTextWindowState);
	procedure AddStat(tx, ty: integer; element: byte; color, tcycle: integer; template: TStat);
	procedure RemoveStat(statId: integer);
	function GetStatIdAt(x, y: integer): integer;
	function BoardPrepareTileForPlacement(x, y: integer): boolean;
	procedure MoveStat(statId: integer; newX, newY: integer);
	procedure PopupPromptString(question: string; var buffer: TString50);
	function Signum(val: integer): integer;
	function Difference(a, b: integer): integer;
	procedure DamageStat(attackerStatId: integer);
	procedure BoardDamageTile(x, y: integer);
	procedure BoardAttack(attackerStatId: integer; x, y: integer);
	procedure CalcDirectionRnd(var deltaX, deltaY: integer);
	procedure CalcDirectionSeek(x, y: integer; var deltaX, deltaY: integer);
	procedure TransitionDrawBoardChange;
	procedure GameUpdateSidebar;
	procedure GameAboutScreen;
	procedure GamePlayLoop(boardChanged: boolean);
	procedure GameDrawSidebar;
	procedure DisplayMessage(time: integer; message: string);
	procedure BoardEnter;
	procedure GameDebugPrompt;
	procedure UseInventoryItem(x,y: integer);
	procedure GameInventoryPrompt;
	procedure DrawInventorySelect(x, y: integer);
	procedure GameTitleLoop;
	procedure HighScoresLoad;
	procedure HighScoresSave;
	procedure HighScoresDisplay(linePos: integer);
	procedure HighScoresAdd(score: integer);
	procedure GameFadeOut(value: integer);
	procedure GameFadeWhite(value: integer);
	procedure GameFadeIn(value: integer);
	procedure GameFadeDown(value: integer);
const
	LFSR11_START = 42;
	ITEM_MAX = 29;
	ProgressAnimColors: array[0 .. 7] of byte = ($14, $1C, $15, $1D, $16, $1E, $17, $1F);
	ProgressAnimStrings: array[0 .. 7] of string[5] =
		('....|', '...*/', '..*.-', '.*..\', '*...|', '..../', '....-', '....\');
	ColorNames: array[1 .. 7] of string[8] =
		('Blue', 'Green', 'Cyan', 'Red', 'Purple', 'Yellow', 'White');
	{}
	ItemNames: array[1..ITEM_MAX] of string[8] =
		('SCONCE','PLANT','MAP','SOCKS','SHIRT','PANTS',
		 'CANDLE','BUCKET','WATER','SHIELD','LADDER','MEAT',
		 'PAINTING','FLASK','PLATTER','MIRROR','LANTERN','GOBLET',
		 'AMETHISP','HELMET','ANVIL','REDWHEN','BLUEWHEN','VIOLETTA',
		 'PLANKS','TOWEL','MILPOR','SCRYER','COIN');

	ItemClass: array[1..ITEM_MAX] of string[8] =
		('LIGHT','FRAGILE','FRAGILE','CLOTHES','CLOTHES','CLOTHES',
		 'LIGHT','FLUID','FLUID','METAL','FRAGILE','FRAGILE',
		 'FRAGILE','FLUID','METAL','FRAGILE','LIGHT','FLUID',
		 'FRAGILE','METAL','METAL','FLUID','FLUID','FRAGILE',
		 'FRAGILE','CLOTHES','FRAGILE','FRAGILE','METAL');
	
	DiagonalDeltaX: array[0 .. 7] of integer = (-1, 0, 1, 1, 1, 0, -1, -1);
	DiagonalDeltaY: array[0 .. 7] of integer = (1, 1, 1, 0, -1, -1, -1, 0);
	NeighborDeltaX: array[0 .. 3] of integer = (0, 0, -1, 1);
	NeighborDeltaY: array[0 .. 3] of integer = (-1, 1, 0, 0);
	{}
	TileBorder: TTile = (Element: E_NORMAL; Color: $0E);
	TileBoardEdge: TTile = (Element: E_BOARD_EDGE; Color: $00);
	StatTemplateDefault: TStat = (
		X: 0; Y: 0; StepX: 0; StepY: 0;
		Cycle: 0; P1: 0; P2: 0; P3: 0;
		Follower: -1; Leader: -1
	);
	LineChars: string[16] = #249#208#210#186#181#188#187#185#198#200#201#204#205#202#203#206;
	VineChars: string[7] = #238#236#237#235#189#190#176;

implementation
uses Dos, Crt, Video, Sounds, Input, Elements, Oop;

function SinusCurve(value: integer): integer;
	var
		stage: integer;
	begin
		stage := value mod 32;
		case stage div 8 of
			0: SinusCurve := stage - 3;
			1: SinusCurve := 6;
			2: SinusCurve := 21 - stage;
		else SinusCurve := -3 end;
	end;


function LFSR11UpdateSeed(var seed: word; var tx, ty: byte): boolean;
	begin
		seed := (seed shr 1) xor ((-(seed and 1)) and $740);
		tx := (seed - 1) and $3F;
		ty := ((seed - 1) shr 6) and $3F;
		LFSR11UpdateSeed := seed = LFSR11_START;
	end;

procedure GameFadeOut(value: integer);
	begin
		FadePalSet(value);
		FadedOut := true;
		FadeTimer := 0;
		repeat
			{TransitionDrawToBoard;}
			{GameDrawSidebar;}
			Inc(FadeTimer);
		until (FadeOut or (FadeTimer > 30));
	end;

procedure GameFadeWhite(value: integer);
	begin
		FadePalSet(value);
		FadedOut := true;
		FadeTimer := 0;
		repeat
			{TransitionDrawToBoard;}
			{GameDrawSidebar;}
			Inc(FadeTimer);
		until (FadeWhite or (FadeTimer > 30));
	end;

procedure GameFadeDown(value: integer);
	begin
		FadeTimer := 0;
		repeat
			{TransitionDrawToBoard;}
			{GameDrawSidebar;}
			Inc(FadeTimer);
		until (FadeDown(value) or (FadeTimer > 30));
		FadedOut := false;
	end;

procedure GameFadeIn(value: integer);
	begin
		FadeTimer := 0;
		repeat
			{TransitionDrawToBoard;}
			{GameDrawSidebar;}
			Inc(FadeTimer);
		until (FadeIn(value) or (FadeTimer > 30));
		FadedOut := false;
	end;

procedure SidebarClearLine(y: integer);
	begin
		VideoWriteText(60, y, $11, #179'                   ');
	end;

procedure SidebarClear;
	var
		i: integer;
	begin
		for i := 3 to 24 do
			SidebarClearLine(i);
	end;

procedure HighScoresLoad;
	var
		f: file of THighScoreList;
		i: integer;
	begin
		Assign(f, 'BINB.HI');
		Reset(f);
		if IOResult = 0 then begin
			Read(f, HighScoreList);
		end;
		Close(f);
		if IOResult <> 0 then begin
			for i := 1 to 30 do begin
				HighScoreList[i].Name := '';
				HighScoreList[i].Score := -1;
			end;
		end;
	end;

procedure HighScoresSave;
	var
		f: file of THighScoreList;
	begin
		Assign(f, 'BINB.HI');
		Rewrite(f);
		Write(f, HighScoreList);
		Close(f);
		if DisplayIOError then begin
		end else begin
		end;
	end;

{$F+}

procedure HighScoresInitTextWindow(var state: TTextWindowState);
	var
		i: integer;
		scoreStr: string;
	begin
		TextWindowInitState(state);
		TextWindowAppend(state, 'Points  Name');
		TextWindowAppend(state, '------  ---------------------------------');
		for i := 1 to HIGH_SCORE_COUNT do begin
			if Length(HighScoreList[i].Name) <> 0 then begin
				Str(HighScoreList[i].Score:5, scoreStr);
				TextWindowAppend(state, scoreStr + '   ' + HighScoreList[i].Name);
			end;
		end;
	end;

procedure HighScoresDisplay(linePos: integer);
	var
		state: TTextWindowState;
	begin
		state.LinePos := linePos;
		HighScoresInitTextWindow(state);
		if (state.LineCount > 2) then begin
			state.Title := 'High points for ' + World.Info.Name;
			TextWindowDrawOpen(state);
			TextWindowSelect(state, false, true);
			TextWindowDrawClose(state);
		end;
		TextWindowFree(state);
	end;

procedure HighScoresAdd(score: integer);
	var
		textWindow: TTextWindowState;
		name: TString50;
		i, listPos: integer;
	begin
		listPos := 1;
		while (listPos <= 30) and (score < HighScoreList[listPos].Score) do
			Inc(listPos);
		if (listPos <= 30) and (score > 0) then begin
			for i := 29 downto listPos do
				HighScoreList[i + 1] := HighScoreList[i];
			HighScoreList[listPos].Score := score;
			HighScoreList[listPos].Name := '-- You! --';

			HighScoresInitTextWindow(textWindow);
			textWindow.LinePos := listPos;
			textWindow.Title := 'New high score for ' + World.Info.Name;
			TextWindowDrawOpen(textWindow);
			TextWindowDraw(textWindow, false, false);

			name := '';
			PopupPromptString('Congratulations!  Enter your name:', name);
			HighScoreList[listPos].Name := name;
			HighScoresSave;

			TextWindowDrawClose(textWindow);
			TransitionDrawToBoard;
			TextWindowFree(textWindow);
		end;
	end;

procedure AdvancePointer(var address: pointer; count: integer);
	begin
		address := Ptr(Seg(address^), Ofs(address^) + count);
	end;

procedure BoardClose;
	var
		ix, iy: integer;
		ptr: pointer;
		rle: TRleTile;
	begin
		if GetStatIdAt(BOARD_WIDTH+1, 0) > 0 then begin
			RemoveStat(GetStatIdAt(BOARD_WIDTH+1, 0));
			if Board.Tiles[BOARD_WIDTH+1][0].Element = E_GLOBAL then
				Board.Tiles[BOARD_WIDTH+1][0].Element := E_BOARD_EDGE;
		end;

		ptr := IoTmpBuf;

		Move(Board.Name, ptr^, SizeOf(Board.Name));
		AdvancePointer(ptr, SizeOf(Board.Name));

		ix := 1;
		iy := 1;
		rle.Count := 1;
		rle.Tile := Board.Tiles[ix][iy];
		repeat
			Inc(ix);
			if ix > BOARD_WIDTH then begin
				ix := 1;
				Inc(iy);
			end;
			if (Board.Tiles[ix][iy].Color = rle.Tile.Color) and
				(Board.Tiles[ix][iy].Element = rle.Tile.Element) and
				(rle.Count < 255) and (iy <= BOARD_HEIGHT) then
			begin
				Inc(rle.Count);
			end else begin
				Move(rle, ptr^, SizeOf(rle));
				AdvancePointer(ptr, SizeOf(rle));
				rle.Tile := Board.Tiles[ix][iy];
				rle.Count := 1;
			end;
		until iy > BOARD_HEIGHT;

		Move(Board.Info, ptr^, SizeOf(Board.Info));
		AdvancePointer(ptr, SizeOf(Board.Info));

		Move(Board.StatCount, ptr^, SizeOf(Board.StatCount));
		AdvancePointer(ptr, SizeOf(Board.StatCount));

		for ix := 0 to Board.StatCount do begin
			with Board.Stats[ix] do begin
				if DataLen > 0 then begin
					for iy := 1 to (ix - 1) do begin
						if Board.Stats[iy].Data = Data then
							DataLen := -iy;
					end;
				end;
				Move(Board.Stats[ix], ptr^, SizeOf(TStat));
				AdvancePointer(ptr, SizeOf(TStat));
				if DataLen > 0 then begin
					Move(Data^, ptr^, DataLen);
					FreeMem(Data, DataLen);
					AdvancePointer(ptr, DataLen);
				end;
			end;
		end;

		FreeMem(World.BoardData[World.Info.CurrentBoard], World.BoardLen[World.Info.CurrentBoard]);
		World.BoardLen[World.Info.CurrentBoard] := Ofs(ptr^) - Ofs(IoTmpBuf^);
		GetMem(World.BoardData[World.Info.CurrentBoard], World.BoardLen[World.Info.CurrentBoard]);
		Move(IoTmpBuf^, World.BoardData[World.Info.CurrentBoard]^, World.BoardLen[World.Info.CurrentBoard]);
	end;

procedure BoardOpen(boardId: integer);
	var
		ptr: pointer;
		ix, iy: integer;
		rle: TRleTile;
	begin
		if boardId > World.BoardCount then
			boardId := World.Info.CurrentBoard;

		ptr := World.BoardData[boardId];

		Move(ptr^, Board.Name, SizeOf(Board.Name));
		AdvancePointer(ptr, SizeOf(Board.Name));

		ix := 1;
		iy := 1;
		rle.Count := 0;
		repeat
			if rle.Count <= 0 then begin
				Move(ptr^, rle, SizeOf(rle));
				AdvancePointer(ptr, SizeOf(rle));
			end;
			Board.Tiles[ix][iy] := rle.Tile;
			Inc(ix);
			if ix > BOARD_WIDTH then begin
				ix := 1;
				Inc(iy);
			end;
			Dec(rle.Count);
		until iy > BOARD_HEIGHT;

		Move(ptr^, Board.Info, SizeOf(Board.Info));
		AdvancePointer(ptr, SizeOf(Board.Info));

		Move(ptr^, Board.StatCount, SizeOf(Board.StatCount));
		AdvancePointer(ptr, SizeOf(Board.StatCount));

		for ix := 0 to Board.StatCount do
			with Board.Stats[ix] do begin
				Move(ptr^, Board.Stats[ix], SizeOf(TStat));
				AdvancePointer(ptr, SizeOf(TStat));
				if DataLen > 0 then begin
					GetMem(Data, DataLen);
					Move(ptr^, Data^, DataLen);
					AdvancePointer(ptr, DataLen);
				end else if DataLen < 0 then begin
					Data := Board.Stats[-DataLen].Data;
					DataLen := Board.Stats[-DataLen].DataLen;
				end;
			end;
		World.Info.CurrentBoard := boardId;
	end;

procedure BoardChange(boardId: integer);
	begin
		Board.Tiles[Board.Stats[0].X][Board.Stats[0].Y].Element := E_PLAYER;
		Board.Tiles[Board.Stats[0].X][Board.Stats[0].Y].Color := World.Info.PlayerColor;
		BoardClose;
		BoardOpen(boardId);
	end;

procedure BoardCreate;
	var
		ix, iy, i: integer;
	begin
		Board.Name := '';
		Board.Info.HintDir := 255;
		Board.Info.IsDark := 0;
		for i := 0 to 3 do
			Board.Info.NeighborBoards[i] := 0;
		Board.Info.ReenterWhenZapped := false;
		Board.Info.Message := '';
		Board.Info.Local := 0;		
		for i := 1 to 8 do
			Board.Info.Register[i] := 0;
		for ix := 0 to BOARD_WIDTH+1 do begin
			Board.Tiles[ix][0] := TileBoardEdge;
			Board.Tiles[ix][BOARD_HEIGHT+1] := TileBoardEdge;
		end;
		for iy := 0 to BOARD_HEIGHT+1 do begin
			Board.Tiles[0][iy] := TileBoardEdge;
			Board.Tiles[BOARD_WIDTH+1][iy] := TileBoardEdge;
		end;

		for ix := 1 to BOARD_WIDTH do
			for iy := 1 to BOARD_HEIGHT do begin
				Board.Tiles[ix][iy].Element := E_EMPTY;
				Board.Tiles[ix][iy].Color := 0;
			end;

		for ix := 1 to BOARD_WIDTH do begin
			Board.Tiles[ix][1] := TileBorder;
			Board.Tiles[ix][BOARD_HEIGHT] := TileBorder;
		end;
		for iy := 1 to BOARD_HEIGHT do begin
			Board.Tiles[1][iy] := TileBorder;
			Board.Tiles[BOARD_WIDTH][iy] := TileBorder;
		end;

		Board.Tiles[BOARD_WIDTH div 2][BOARD_HEIGHT div 2].Element := E_PLAYER;
		Board.Tiles[BOARD_WIDTH div 2][BOARD_HEIGHT div 2].Color := World.Info.PlayerColor;
		Board.StatCount := 0;
		Board.Stats[0].X := BOARD_WIDTH div 2;
		Board.Stats[0].Y := BOARD_HEIGHT div 2;
		Board.Stats[0].Cycle := 1;
		Board.Stats[0].Under.Element := E_EMPTY;
		Board.Stats[0].Under.Color := 0;
		Board.Stats[0].Data := nil;
		Board.Stats[0].DataLen := 0;
	end;

procedure WorldCreate;
	var
		i: integer;
	begin
		InitElementsGame;
		World.BoardCount := 0;
		World.BoardLen[0] := 0;
		BoardCreate;
		World.Info.CurrentBoard := 0;
		World.Info.Ammo := 0;
		World.Info.Gems := 0;
		World.Info.Health := 100;
		World.Info.RomBank := 0;
		World.Info.CustomZ := 0;
		World.Info.Torches := 0;
		World.Info.TorchTicks := 0;
		World.Info.Score := 0;
		for i := 1 to 6 do
			World.Info.Keys[i] := 97;
		for i := 1 to 10 do
			World.Info.Register[i] := 0;
		for i := 1 to 25 do
			World.Info.Flags[i] := '';
		if ConfigPaper then begin
			World.Info.PlayerColor := $09;
		end else begin
			World.Info.PlayerColor := $1F;
		end;
		BoardChange(0);
		Board.Name := 'Title screen';
		World.Info.Name := '';
	end;

procedure TransitionDrawToFill(chr: char; color: integer);
	var
		seed: word;
		tx, ty: byte;
	begin
		seed := LFSR11_START;
		repeat
			if (tx < 60) and (ty < 25) then
				VideoWriteText(tx, ty, color, chr);
		until LFSR11UpdateSeed(seed, tx, ty);

		if (tx < 60) and (ty < 25) then
			VideoWriteText(tx, ty, color, chr);
	end;

procedure OopLaserBeam(x, y, dx, dy: integer);
	var
		tile: TTile;
	begin
		VideoWriteText(x - 1, y - 1, $0F, #177);
		tile := Board.Tiles[x + dx][y + dy];
		if ((tile.Element = E_EMPTY) or (tile.Element = E_TEXT_BLUE)) or ((tile.Element = E_OBJECT) and (tile.Color = $09)) then
			OopLaserBeam(x + dx, y + dy, dx, dy)
		else begin
			if tile.Element = E_OBJECT then begin
				if tile.Color = $97 then begin
					if Board.Tiles[x + dx + dy][y + dx + dy].Element <> E_PLAYER then
						OopLaserBeam(x + dx + dy, y + dx + dy, dy, dx);			
				end;
				if tile.Color = $9F then begin
					if Board.Tiles[x + dx - dy][y + dy - dx].Element <> E_PLAYER then
						OopLaserBeam(x + dx - dy, y + dy - dx, -dy, -dx);
				end;
				if tile.Color = $08 then begin
					if OopSend(0,'ALL:LASERWIN', false) then begin end;
				end;
			end;
		end;
	end;

procedure BoardDrawTile(x, y: integer);
	var
		ch: byte;
		cl: byte;
	begin
		with Board.Tiles[x][y] do begin
			if (Board.Info.IsDark <> 1)
				or (ElementDefs[Board.Tiles[x][y].Element].VisibleInDark)
				or (
					(World.Info.TorchTicks > 0)
					and ((Sqr(Board.Stats[0].X - x) + Sqr(Board.Stats[0].Y - y) * 7 / 4) <=
						((World.Info.TorchTicks div 8) + (SinusCurve(CurrentTick))))
				) or ForceDarknessOff then
			begin
				if Element = E_EMPTY then begin
					if ConfigPaper then begin
						VideoWriteText(x - 1, y - 1, $F0, ' ');
					end else begin
						VideoWriteText(x - 1, y - 1, $0F, ' ');
					end;
				end else if ElementDefs[Element].HasDrawProc then begin
					if (GetStatIdAt(x, y) >= 0) and (Board.Stats[GetStatIdAt(x, y)].Under.Element = E_OVERLAY) then begin
						VideoWriteText(x - 1, y - 1, Board.Stats[GetStatIdAt(x, y)].Under.Color, #176);
					end else begin
						ElementDefs[Element].DrawProc(x, y, ch);
						VideoWriteText(x - 1, y - 1, Color, Chr(ch));
					end;
				end else if (Element < E_TEXT_MIN) or (Element > E_TEXT_MAX) then begin
					if (GetStatIdAt(x, y) >= 0) and (Board.Stats[GetStatIdAt(x, y)].Under.Element = E_OVERLAY) then
						VideoWriteText(x - 1, y - 1, Board.Stats[GetStatIdAt(x, y)].Under.Color, #176)
					else
						VideoWriteText(x - 1, y - 1, Color, ElementDefs[Element].Character);
				end else begin
					{ Text drawing }
					case Board.Tiles[x][y].Color of
						$00, $20, $FF: Board.Tiles[x][y].Color := $20;
					end;
					if Element = E_TEXT_WHITE then
						VideoWriteText(x - 1, y - 1, $F0, Chr(Board.Tiles[x][y].Color))
					else if Element = E_TEXT_BLUE then
						VideoWriteText(x - 1, y - 1, $09, Chr(Board.Tiles[x][y].Color))
					else if Element = E_TEXT_PURPLE then
						VideoWriteText(x - 1, y - 1, $08, Chr(Board.Tiles[x][y].Color))
					else
						VideoWriteText(x - 1, y - 1, ((Element - E_TEXT_MIN) + 1), Chr(Board.Tiles[x][y].Color));
				end;
			end else begin
				{ Darkness }
				{if (Color and $0F > $09) and ((Sqr(Board.Stats[0].X - x) + Sqr(Board.Stats[0].Y - y) * 7 / 4) <= 25) then
					VideoWriteText(x - 1, y - 1, (Color and $0F) - $08, #198)
				else if (Color and $0F > 0) and ((Sqr(Board.Stats[0].X - x) + Sqr(Board.Stats[0].Y - y) * 7 / 4) <= 25) then
					VideoWriteText(x - 1, y - 1, Color and $0F, #198)
				else}
					VideoWriteText(x - 1, y - 1, $08, #198);
			end;
		end;
	end;

procedure BoardDrawBorder;
	var
		ix, iy: integer;
	begin
		for ix := 1 to BOARD_WIDTH do begin
			BoardDrawTile(ix, 1);
			BoardDrawTile(ix, BOARD_HEIGHT);
		end;

		for iy := 1 to BOARD_HEIGHT do begin
			BoardDrawTile(1, iy);
			BoardDrawTile(BOARD_WIDTH, iy);
		end;
	end;

procedure TransitionDrawToBoard;
	var
		seed: word;
		tx, ty: byte;
	begin
		if not FadedOut then begin
			if Board.Info.IsDark < 2 then
				SetPal(1)
			else
				SetPal(Board.Info.IsDark);
		end;
		BoardDrawBorder;

		seed := LFSR11_START;
		repeat
			if (tx < 60) and (ty < 25) then
				BoardDrawTile(tx+1, ty+1);
		until LFSR11UpdateSeed(seed, tx, ty);


		if (tx < 60) and (ty < 25) then
			BoardDrawTile(tx+1, ty+1);
	end;

procedure SidebarPromptSlider(editable: boolean; x, y: integer; prompt: string; var value: byte);
	var
		newValue: integer;
		startChar, endChar: char;
	begin
		if prompt[Length(prompt) - 2] = ';' then begin
			startChar := prompt[Length(prompt) - 1];
			endChar := prompt[Length(prompt)];
			prompt := Copy(prompt, 1, Length(prompt) - 3);
		end else begin
			startChar := '1';
			endChar := '9';
		end;

		SidebarClearLine(y);
		VideoWriteText(x, y, Integer(editable) + $1E, prompt);
		SidebarClearLine(y + 1);
		SidebarClearLine(y + 2);
		VideoWriteText(x, y + 2, $1e, startChar + '....:....' + endChar);

		SlowInput;
		repeat
			if editable then begin
				if InputJoystickMoved then
					TrueDelay(45);
				VideoWriteText(x + value + 1, y + 1, $9F, #31);

				InputUpdate;
				if (InputKeyPressed >= '1') and (InputKeyPressed <= '9') then begin
					value := Ord(InputKeyPressed) - 49;
					SidebarClearLine(y + 1);
				end else begin
					newValue := value + InputDeltaX;
					if (value <> newValue) and (newValue >= 0) and (newValue <= 8) then begin
						value := newValue;
						SidebarClearLine(y + 1);
					end;
				end;
			end;
		until (InputKeyPressed = KEY_ENTER) or (InputKeyPressed = KEY_ESCAPE) or not editable;
		{FastInput;}

		VideoWriteText(x + value + 1, y + 1, $1F, #31);
	end;

procedure PromptString(x, y, arrowColor, color, width: integer; mode: byte; var buffer: TString50);
	var
		i: integer;
		oldBuffer: string;
	begin
		oldBuffer := buffer;

		TrueDelay(75);
		SlowInput;
		repeat
			for i := 0 to (width - 1) do begin
				VideoWriteText(x + i, y, color, ' ');
				VideoWriteText(x + i, y - 1, arrowColor, ' ');
			end;
			VideoWriteText(x + width, y - 1, arrowColor, ' ');
			VideoWriteText(x + Length(buffer), y - 1, (arrowColor div $10) * 16 + $0F, #31);
			VideoWriteText(x, y, color, buffer);

			InputReadWaitKey;

			if (Length(buffer) < width) and (InputKeyPressed >= #32) and (InputKeyPressed < #128) then begin
				case mode of
					PROMPT_NUMERIC: begin
						if (InputKeyPressed in ['0' .. '9']) then begin
							buffer := buffer + InputKeyPressed;
						end;
					end;
					PROMPT_ANY: begin
						if (InputKeyPressed <> #0) and (InputKeyPressed <> '?') then
							buffer := buffer + InputKeyPressed;
					end;
					PROMPT_ALPHANUM: begin
						if (UpCase(InputKeyPressed) in ['A' .. 'Z'])
							or (InputKeyPressed in ['0' .. '9'])
							or (InputKeyPressed = '-') then
						begin
							buffer := buffer + UpCase(InputKeyPressed);
						end;
					end;
				end;
			end else if (InputKeyPressed = KEY_LEFT) or (InputKeyPressed = KEY_BACKSPACE) then begin
				buffer := Copy(buffer, 1, Length(buffer) - 1);
			end;

		until ((InputKeyPressed = KEY_ENTER) and (InputDeltaY = 0)) or (InputKeyPressed = KEY_ESCAPE);
		{FastInput;}
		if InputKeyPressed = KEY_ESCAPE then begin
			buffer := oldBuffer;
		end;
	end;

function SidebarPromptYesNo(message: string; defaultReturn: boolean): boolean;
	begin
		SidebarClearLine(3);
		SidebarClearLine(4);
		SidebarClearLine(5);
		VideoWriteText(63, 5, $1F, message);
		VideoWriteText(63 + Length(message), 5, $9E, '_');

		SlowInput;
		repeat
			InputReadWaitKey;
		until UpCase(InputKeyPressed) in ['N', 'Y'];
			if UpCase(InputKeyPressed) = 'Y' then begin
				defaultReturn := true;
			end else
				defaultReturn := false;

		SidebarClearLine(5);
		SidebarPromptYesNo := defaultReturn;
		{FastInput;}
	end;

procedure SidebarPromptString(prompt: string; extension: TString50; var filename: string; promptMode: byte);
	begin
		SidebarClearLine(3);
		SidebarClearLine(4);
		SidebarClearLine(5);
		VideoWriteText(75 - Length(prompt), 3, $1F, prompt);
		VideoWriteText(63, 5, $0F, '        ' + extension);

		PromptString(63, 5, $1E, $0F, 8, promptMode, filename);

		SidebarClearLine(3);
		SidebarClearLine(4);
		SidebarClearLine(5);
	end;

procedure PauseOnError;
	begin
		SoundQueue(1, SoundParse('004x114x9'));
		TrueDelay(100);
	end;

function DisplayIOError: boolean;
	var
		errorNumStr: TString50;
		textWindow: TTextWindowState;
	begin
		if IOResult = 0 then begin
			DisplayIOError := false;
			exit;
		end;

		DisplayIOError := true;
	end;

procedure WorldUnload;
	var
		i: integer;
	begin
		BoardClose;
		for i := 0 to World.BoardCount do
			FreeMem(World.BoardData[i], World.BoardLen[i]);
	end;

procedure WorldSwap(bank, brd: integer);
	var
		f: file;
		ptr: pointer;
		boardId: integer;
		fn: TString50;
		i: integer;
		state: TTextWindowState;
	begin
		{Save Current World}
		SwapWorld := World.Info;
		WorldSave(SwapWorld.RomBank);
		{Load Swap World}
		Str(bank, fn);
		i := 0;
		while i < 5 do begin
			if WorldLoad('BINBSAVE', '.S' + fn) then
				i := 10
			else
				Inc(i);
		end;
		if (i < 10) then begin
			if WorldLoad('BINB' + fn, '.ZZT') then begin
				GameFadeIn(2);
				TextWindowInitState(state);
				TextWindowAppend(state, 'CREATING SAVE FILE CHUNK #' + fn);
				TextWindowAppend(state, '---------------------------------');
				TextWindowAppend(state, 'We cannot seem to find chunk #' + fn);
				TextWindowAppend(state, 'of your save file.  We are creating');
				TextWindowAppend(state, 'a new one now. If you force close');
				TextWindowAppend(state, 'this window now, new data will not be');
				TextWindowAppend(state, 'created and nothing will be overwritten.');
				state.Title := 'CREATING SAVE FILE CHUNK #' + fn;
				TextWindowDrawOpen(state);
				TextWindowSelect(state, false, false);
				TextWindowDrawClose(state);
				TextWindowFree(state);
			end;
		end;

		BoardClose;
		World.Info := SwapWorld;
		World.Info.RomBank := bank;
		World.Info.CurrentBoard := brd;
		BoardOpen(World.Info.CurrentBoard);
	end;

function WorldTest(bank: integer): boolean;
	var
		f: file;
		fn: string[1];
	begin
		WorldTest := false;
		Str(bank, fn);
		Assign(f, 'BINB' + fn + '.ZZT');
		Reset(f, 1);
		if not DisplayIOError then begin
			WorldTest := true;
			Close(f);
		end;
	end;

function SaveTest: boolean;
	var
		f: file;
	begin
		SaveTest := false;
		Assign(f, 'BINBSAVE.S99');
		Reset(f, 1);
		if not DisplayIOError then begin
			SaveTest := true;
			Close(f);
		end;
	end;


function WorldLoad(filename, extension: TString50): boolean;
	var
		f: file;
		ptr: pointer;
		boardId: integer;
	begin
		WorldLoad := false;
{ ## GLOBAL OBJECT }
		GlobalRead := false;
{ ## END GLOBAL OBJECT }

		Assign(f, filename + extension);
		Reset(f, 1);

		if not DisplayIOError then begin
			WorldUnload;
			BlockRead(f, IoTmpBuf^, WORLD_FILE_HEADER_SIZE);

			if not DisplayIOError then begin
				ptr := IoTmpBuf;
				Move(ptr^, World.BoardCount, SizeOf(World.BoardCount));
				AdvancePointer(ptr, SizeOf(World.BoardCount));

				if World.BoardCount < 0 then begin
					if World.BoardCount <> -1 then begin
						VideoWriteText(63, 5, $1E, 'You need a newer');
						VideoWriteText(63, 6, $1E, ' version of ZZT!');
						exit;
					end else begin
						Move(ptr^, World.BoardCount, SizeOf(World.BoardCount));
						AdvancePointer(ptr, SizeOf(World.BoardCount));
					end;
				end;

				Move(ptr^, World.Info, SizeOf(World.Info));
				AdvancePointer(ptr, SizeOf(World.Info));

{ ## GLOBAL OBJECT }
				BlockRead(f, World.BoardLen[0], 2);
				GetMem(World.BoardData[0], World.BoardLen[0]);
				BlockRead(f, World.BoardData[0]^, World.BoardLen[0]);
				boardId := World.Info.CurrentBoard;
				BoardOpen(0);

				if Board.Tiles[1][1].Element = E_OBJECT then begin
					with Board.Stats[GetStatIdAt(1,1)] do begin
						if DataLen > GlobalLen then begin
							New(GlobalData);
							GlobalLen := DataLen;
							GetMem(GlobalData, GlobalLen);
							Move(Data^, GlobalData^, GlobalLen);
						end;
					end;
					Board.Tiles[1][1].Element := E_EMPTY;
				end;
				BoardClose;
				World.Info.CurrentBoard := boardId;
				if World.BoardCount > 0 then begin
					for boardId := 1 to World.BoardCount do begin
						BlockRead(f, World.BoardLen[boardId], 2);
						GetMem(World.BoardData[boardId], World.BoardLen[boardId]);
						BlockRead(f, World.BoardData[boardId]^, World.BoardLen[boardId]);
					end;
				end;
{ ## END GLOBAL OBJECT }				

				Close(f);

				BoardOpen(World.Info.CurrentBoard);
				WorldLoad := true;

				HighScoresLoad;

			end;
		end;
	end;

procedure WorldSave(bank: integer);
	var
		f: file;
		i, j, k: integer;
		fn: TString50;
		ptr: pointer;
		version: integer;
		state: TTextWindowState;
	begin
		BoardClose;
		Str(bank, fn);

		j := 0;
		while (j < 5) do begin
			Assign(f, 'BINBSAVE.S' + fn);
			Rewrite(f, 1);
			if not DisplayIOError then
				j := 10
			else
				Inc(j);
				TrueDelay(200);
		end;

		if (j = 10) then begin
			ptr := IoTmpBuf;
			FillChar(IoTmpBuf^, WORLD_FILE_HEADER_SIZE, 0);
			version := -1;
			Move(version, ptr^, SizeOf(version));
			AdvancePointer(ptr, SizeOf(version));

			Move(World.BoardCount, ptr^, SizeOf(World.BoardCount));
			AdvancePointer(ptr, SizeOf(World.BoardCount));

			Move(World.Info, ptr^, SizeOf(World.Info));
			AdvancePointer(ptr, SizeOf(World.Info));

			k := 0;
			while (k < 5) do begin
				BlockWrite(f, IoTmpBuf^, WORLD_FILE_HEADER_SIZE);
				if not DisplayIOError then
					k := 10
				else
					Inc(k);
					TrueDelay(200);
			end;
			if (k < 10) then begin
				GameFadeIn(2);
				TextWindowInitState(state);
				TextWindowAppend(state, 'CANNOT SAVE FILE CHUNK #' + fn);
				TextWindowAppend(state, '---------------------------------');
				TextWindowAppend(state, 'We cannot seem to save chunk #' + fn);
				TextWindowAppend(state, 'of your save file. Unfortunately,');
				TextWindowAppend(state, 'This probably means data has been');
				TextWindowAppend(state, 'lost. Believe me, I hate this more');
				TextWindowAppend(state, 'than you do.');
				state.Title := 'CANNOT SAVE FILE CHUNK #' + fn;
				TextWindowDrawOpen(state);
				TextWindowSelect(state, false, false);
				TextWindowDrawClose(state);
				TextWindowFree(state);
			end;
		end else begin
			GameFadeIn(2);
			TextWindowInitState(state);
			TextWindowAppend(state, 'CANNOT OPEN FILE CHUNK #' + fn);
			TextWindowAppend(state, '---------------------------------');
			TextWindowAppend(state, 'We cannot seem to open chunk #' + fn);
			TextWindowAppend(state, 'of your save file for saving.');
			TextWindowAppend(state, 'Regrettably, data has been lost. :(');
			state.Title := 'CANNOT OPEN FILE CHUNK #' + fn;
			TextWindowDrawOpen(state);
			TextWindowSelect(state, false, false);
			TextWindowDrawClose(state);
			TextWindowFree(state);
		end;

		for i := 0 to World.BoardCount do begin
			BlockWrite(f, World.BoardLen[i], 2);
			BlockWrite(f, World.BoardData[i]^, World.BoardLen[i]);
		end;

		BoardOpen(World.Info.CurrentBoard);
		Close(f);
		exit;

	end;

procedure CopyStatDataToTextWindow(statId: integer; var state: TTextWindowState);
	var
		dataStr: string;
		dataPtr: pointer;
		dataChr: char;
		i: integer;
	begin
		with Board.Stats[statId] do begin
			TextWindowInitState(state);
			dataStr := '';
			dataPtr := Data;

			for i := 0 to DataLen do begin
				Move(dataPtr^, dataChr, 1);
				if dataChr = KEY_ENTER then begin
					TextWindowAppend(state, dataStr);
					dataStr := '';
				end else begin
					dataStr := dataStr + dataChr;
				end;
				AdvancePointer(dataPtr, 1);
			end;
		end;
	end;

procedure AddStat(tx, ty: integer; element: byte; color, tcycle: integer; template: TStat);
	begin
		if Board.StatCount < MAX_STAT then begin
			Inc(Board.StatCount);
			Board.Stats[Board.StatCount] := template;
			with Board.Stats[Board.StatCount] do begin
				X := tx;
				Y := ty;
				Cycle := tcycle;
				Under := Board.Tiles[tx][ty];
				DataPos := 0;
				DataLast := 0;
				MyLocal[1] := 0;
				MyLocal[2] := 0;
				MyLocal[3] := 0;
			end;

			if template.Data <> nil then begin
				GetMem(Board.Stats[Board.StatCount].Data, template.DataLen);
				Move(template.Data^, Board.Stats[Board.StatCount].Data^, template.DataLen);
			end;

			if ElementDefs[Board.Tiles[tx][ty].Element].PlaceableOnTop then
				Board.Tiles[tx][ty].Color := (color and $0F) + (Board.Tiles[tx][ty].Color and $F0)
			else
				Board.Tiles[tx][ty].Color := color;
			Board.Tiles[tx][ty].Element := element;

			if ty > 0 then
				BoardDrawTile(tx, ty);
		end;
	end;

procedure RemoveStat(statId: integer);
	var
		i: integer;
	label StatDataInUse;
	begin
		with Board.Stats[statId] do begin
			if DataLen <> 0 then begin
				for i := 1 to Board.StatCount do begin
					if (Board.Stats[i].Data = Data) and (i <> statId) then
						goto StatDataInUse;
				end;
				FreeMem(Data, DataLen);
			end;

		StatDataInUse:
			if statId < CurrentStatTicked then
				Dec(CurrentStatTicked);

			Board.Tiles[X][Y] := Under;
			if Y > 0 then
				BoardDrawTile(X, Y);

			for i := 1 to Board.StatCount do begin
				if Board.Stats[i].Follower >= statId then begin
					if Board.Stats[i].Follower = statId then
						Board.Stats[i].Follower := -1
					else
						Dec(Board.Stats[i].Follower);
				end;

				if Board.Stats[i].Leader >= statId then begin
					if Board.Stats[i].Leader = statId then
						Board.Stats[i].Leader := -1
					else
						Dec(Board.Stats[i].Leader);
				end;
			end;

			for i := (statId + 1) to Board.StatCount do
				Board.Stats[i - 1] := Board.Stats[i];
			Dec(Board.StatCount);
		end;
	end;

function GetStatIdAt(x, y: integer): integer;
	var
		i: integer;
	begin
		i := -1;
		repeat
			Inc(i);
		until ((Board.Stats[i].X = x) and (Board.Stats[i].Y = y)) or (i > Board.StatCount);

		if i > Board.StatCount then
			GetStatIdAt := -1
		else
			GetStatIdAt := i;
	end;

function BoardPrepareTileForPlacement(x, y: integer): boolean;
	var
		statId: integer;
		result: boolean;
	begin
		statId := GetStatIdAt(x, y);
		if statId > 0 then begin
			RemoveStat(statId);
			result := true;
		end else if statId < 0 then begin
			if not ElementDefs[Board.Tiles[x][y].Element].PlaceableOnTop then
				Board.Tiles[x][y].Element := E_EMPTY;
			result := true;
		end else begin { statId = 0 (player) cannot be modified }
			result := false;
		end;
		BoardDrawTile(x, y);
		BoardPrepareTileForPlacement := result;
	end;

procedure MoveStat(statId: integer; newX, newY: integer);
	var
		iUnder: TTile;
		ix, iy: integer;
		oldX, oldY: integer;
		oldBgColor: integer;
	begin
		with Board.Stats[statId] do begin
			oldBgColor := Board.Tiles[newX][newY].Color and $F0;

			iUnder := Board.Stats[statId].Under;
			Board.Stats[statId].Under := Board.Tiles[newX][newY];

			if Board.Tiles[X][Y].Element = E_PLAYER then
				Board.Tiles[newX][newY].Color := Board.Tiles[X][Y].Color
			else if Board.Tiles[newX][newY].Element = E_EMPTY then
				Board.Tiles[newX][newY].Color := Board.Tiles[X][Y].Color and $0F
			else if (Board.Tiles[newX][newY].Element < E_TEXT_MIN) or (Board.Tiles[newX][newY].Element > E_TEXT_MAX) then
				Board.Tiles[newX][newY].Color := (Board.Tiles[X][Y].Color and $0F) + (Board.Tiles[newX][newY].Color and $F0)
			else if Board.Tiles[newX][newY].Element = E_TEXT_BLUE then
				Board.Tiles[newX][newY].Color := (Board.Tiles[X][Y].Color and $0F) + $90
			else if Board.Tiles[newX][newY].Element = E_TEXT_WHITE then
				Board.Tiles[newX][newY].Color := (Board.Tiles[X][Y].Color and $0F) + $F0
			else if Board.Tiles[newX][newY].Element = E_TEXT_PURPLE then
				Board.Tiles[newX][newY].Color := (Board.Tiles[X][Y].Color and $0F) + $80
			else
				Board.Tiles[newX][newY].Color := (Board.Tiles[X][Y].Color and $0F) + ((Board.Tiles[newX][newY].Element - 46) * 16);

			Board.Tiles[newX][newY].Element := Board.Tiles[X][Y].Element;
			Board.Tiles[X][Y] := iUnder;

			oldX := X;
			oldY := Y;
			X := newX;
			Y := newY;

			BoardDrawTile(X, Y);
			BoardDrawTile(oldX, oldY);

			{ come into doors already open }
			if Board.Stats[statId].Under.Element = E_BIG_DOOR then begin
				Board.Stats[statId].Under.Element := E_BDUNDER;
				if Board.Tiles[X + 1][Y].Element = E_BIG_DOOR then begin
					Board.Tiles[X + 1][Y].Element := E_BDUNDER;
					BoardDrawTile(X + 1, Y);
				end;
				if Board.Tiles[X][Y + 1].Element = E_BIG_DOOR then begin
					Board.Tiles[X][Y + 1].Element := E_BDUNDER;
					BoardDrawTile(X, Y + 1);
				end;
				if Board.Tiles[X + 1][Y + 1].Element = E_BIG_DOOR then begin
					Board.Tiles[X + 1][Y + 1].Element := E_BDUNDER;
					BoardDrawTile(X + 1, Y + 1);
				end;
				if Board.Tiles[X - 1][Y].Element = E_BIG_DOOR then begin
					Board.Tiles[X - 1][Y].Element := E_BDUNDER;
					BoardDrawTile(X - 1, Y);
				end;
				if Board.Tiles[X][Y - 1].Element = E_BIG_DOOR then begin
					Board.Tiles[X][Y - 1].Element := E_BDUNDER;
					BoardDrawTile(X, Y - 1);
				end;
				if Board.Tiles[X - 1][Y - 1].Element = E_BIG_DOOR then begin
					Board.Tiles[X - 1][Y - 1].Element := E_BDUNDER;
					BoardDrawTile(X - 1, Y - 1);
				end;
				if Board.Tiles[X - 1][Y + 1].Element = E_BIG_DOOR then begin
					Board.Tiles[X - 1][Y + 1].Element := E_BDUNDER;
					BoardDrawTile(X - 1, Y + 1);
				end;
				if Board.Tiles[X + 1][Y - 1].Element = E_BIG_DOOR then begin
					Board.Tiles[X + 1][Y - 1].Element := E_BDUNDER;
					BoardDrawTile(X + 1, Y - 1);
				end;
			end;


			if (statId = 0) and ((Board.Info.IsDark = 1) and (World.Info.TorchTicks > 0)) then begin
				if (Sqr(oldX-X) + Sqr(oldY-Y)) > 1 then
					DrawPlayerSurroundings(oldX, oldY);
				DrawPlayerSurroundings(X, Y);
			end;

		end;
	end;

procedure PopupPromptString(question: string; var buffer: TString50);
	var
		x, y: integer;
	begin
		VideoWriteText(3, 18, $4F, TextWindowStrTop);
		VideoWriteText(3, 19, $4F, TextWindowStrText);
		VideoWriteText(3, 20, $4F, TextWindowStrSep);
		VideoWriteText(3, 21, $4F, TextWindowStrText);
		VideoWriteText(3, 22, $4F, TextWindowStrText);
		VideoWriteText(3, 23, $4F, TextWindowStrBottom);
		VideoWriteText(4 + (TextWindowWidth - Length(question)) div 2, 19, $4F, question);
		buffer := '';
		PromptString(10, 22, $4F, $4E, TextWindowWidth - 16, PROMPT_ANY, buffer);
		for y := 18 to 23 do
			for x := 3 to (TextWindowWidth + 3) do
				BoardDrawTile(x + 1, y + 1);
	end;

function Signum(val: integer): integer;
	begin
		if val > 0 then
			Signum := 1
		else if val < 0 then
			Signum := -1
		else
			Signum := 0;
	end;

function Difference(a, b: integer): integer;
	begin
		if (a - b) >= 0 then
			Difference := a - b
		else
			Difference := b - a;
	end;

procedure DisplayGameClock(clockval, dispval: integer);
	var
		numHour: string[1];
		numTMin: string[1];
		numOMin: string[1];
		numTsec: string[1];
		numOSec: string[1];
	begin
		if dispval > 0 then begin
			Str(clockval div 3600, numHour);
			Str((clockval mod 3600) div 600, numTMin);
			Str((clockval mod 600) div 60, numOMin);
			Str((clockval mod 60) div 10, numTSec);
			Str(clockval mod 10, numOSec);
			VideoWriteText(62, 3, $1B, 'TIME:');
			VideoWriteText(68, 3, $1B, numHour + ':' + numTMin + numOMin + ':' + numTSec + numOSec);
		end else
			VideoWriteText(62, 3, $1F, '               ');
	end;

procedure GameUpdateSidebar;
	var
		numStr: string[8];
		i: integer;
	begin
		if GameStateElement = E_PLAYER then begin

			if World.Info.Health < 0 then
				World.Info.Health := 0;

			Str(World.Info.Health, numStr);
			VideoWriteText(72, 9, $1E, numStr + '% ');

			if FlashInv > 0 then
				VideoWriteText(65, 16, $10 + ((FlashInv mod 2) * 15), ' Inventory')
			else
				VideoWriteText(65, 16, $1F, ' Inventory');

			{ % Completion }
			DisplayGameClock(World.Info.Register[8], DisplayClock);
			{ Stickers are Gems }
			Str(World.Info.Gems, numStr);
			VideoWriteText(72, 10, $1E, numStr + ' ');
			{ Points are Score }
			Str(World.Info.Score, numStr);
			VideoWriteText(72, 11, $1E, numStr + ' ');
			Str(World.Info.CustomZ, numStr);
			if World.Info.CustomZ > 0 then begin
				VideoWriteText(65, 8, $1C, 'DANGER:');
				VideoWriteText(72, 8, $1C, numStr + ' ');
			end else begin
				VideoWriteText(65, 8, $1C, '            ')
			end;
			for i := 1 to 7 do begin
				if World.Info.Register[7] < i then
					VideoWriteText(71 + i, 12, $5D, #176)
				else
					VideoWriteText(71 + i, 12, $05, #177);
			end;
			if World.Info.Torches > 2 then begin
				VideoWriteText(68, 13, $1B, 'Ice:');
				for i := 0 to 5 do begin
					if World.Info.Torches > ((100 * i) + 50) then
						VideoWriteText(72 + i, 13, $3B, #178)
					else
						VideoWriteText(72 + i, 13, $3B, #176);
				end;
			end;
			if World.Info.Ammo > 5 then begin
				VideoWriteText(68, 13, $1C, 'Hot:');
				for i := 0 to 5 do begin
					if World.Info.Ammo > ((100 * i) + 50) then
						VideoWriteText(72 + i, 13, $4C, #178)
					else
						VideoWriteText(72 + i, 13, $4C, #176);
				end;
			end;
			If (World.Info.Ammo < 6) and (World.Info.Torches < 3) then
				VideoWriteText(68, 13, $1E, '           ');
			if (((WorldGetFlagPosition('CANDLE') >= 0) or (WorldGetFlagPosition('SCONCE') >= 0)) or
			   (WorldGetFlagPosition('LANTERN') >= 0)) and (Board.Info.IsDark = 1) then begin
				VideoWriteText(62, 19, $30, ' T ');
				VideoWriteText(65, 19, $1F, ' Light ');
				if (World.Info.TorchTicks < 55) and (World.Info.TorchTicks > 0) then
					i := (World.Info.TorchTicks mod 2) * 15
				else
					i := 0;
				if WorldGetFlagPosition('CANDLE') >= 0 then begin
					VideoWriteText(72, 19, $1F - i, 'Candle');
				end;
				if WorldGetFlagPosition('SCONCE') >= 0 then begin
					VideoWriteText(72, 19, $1F - i, 'Sconce');
				end;
				if WorldGetFlagPosition('LANTERN') >= 0 then begin
					VideoWriteText(72, 19, $1F - i, 'Lantern');
				end;
			end else begin
				SidebarClearLine(19);
			end;
			if WorldGetFlagPosition('AMETHISP') >= 0 then
				if World.Info.Register[1] > 0 then
					VideoWriteText(64, 14, $E0, 'Discount Mode')
				else
					VideoWriteText(64, 14, $13, 'Discount Mode')
			else
				VideoWriteText(64, 14, $10, '             ');

			{if World.Info.TorchTicks = 0 then
				VideoWriteText(75, 9, $16, '    ')
			else begin
				for i := 2 to 5 do begin
					if i <= ((World.Info.TorchTicks * 5) div TORCH_DURATION) then
						VideoWriteText(73 + i, 9, $16, #177)
					else
						VideoWriteText(73 + i, 9, $16, #176);
				end;
			end;}

			if SoundEnabled then
				VideoWriteText(65, 17, $1F, ' Be quiet')
			else
				VideoWriteText(65, 17, $1F, ' Be noisy');
			if DebugEnabled then begin
				Str(MemAvail, numStr);
				VideoWriteText(69, 4, $1E, 'm' + numStr + ' ');
			end;
		end;
	end;

procedure DisplayMessage(time: integer; message: string);
	begin
		if GetStatIdAt(0, 0) <> -1 then begin
			RemoveStat(GetStatIdAt(0, 0));
			BoardDrawBorder;
		end;

		if Length(message) <> 0 then begin
			AddStat(0, 0, E_MESSAGE_TIMER, 0, 1, StatTemplateDefault);
			Board.Stats[Board.StatCount].P2 := Time div (TickTimeDuration + 1);
			Board.Info.Message := message;
		end;
	end;

procedure DamageStat(attackerStatId: integer);
	var
		oldX, oldY: integer;
	begin
		with Board.Stats[attackerStatId] do begin
			if attackerStatId = 0 then begin
				if World.Info.Health > 0 then begin
					World.Info.Health := World.Info.Health - 10;

					GameUpdateSidebar;

					Board.Tiles[X][Y].Color := $4C;

					if World.Info.Health > 0 then begin
						if Board.Info.ReenterWhenZapped then begin
							SoundQueue(4, #32#1#35#1#39#1#48#1#16#1);

							{ Move player to start }
							Board.Tiles[X][Y].Element := E_EMPTY;
							BoardDrawTile(X, Y);
							oldX := X;
							oldY := Y;
							X := Board.Info.StartPlayerX;
							Y := Board.Info.StartPlayerY;
							DrawPlayerSurroundings(oldX, oldY);
							DrawPlayerSurroundings(X, Y);

							GamePaused := true;
						end;
						SoundQueue(4, #16#1#32#1);
					end else begin
						SoundQueue(5, #32#3#35#3#39#3#48#3);
					end;
				end;
			end else begin
				case Board.Tiles[X][Y].Element of
					E_BULLET: SoundQueue(3, SoundParse('7'));
					E_OBJECT: begin end;
				else
					SoundQueue(3, SoundParse('0101'));
				end;
				RemoveStat(attackerStatId);
			end;
		end;
	end;

procedure BoardDamageTile(x, y: integer);
	var
		statId: integer;
	begin
		statId := GetStatIdAt(x, y);
		if statId <> -1 then begin
			DamageStat(statId);
		end else begin
			Board.Tiles[x][y].Element := E_EMPTY;
			BoardDrawTile(x, y);
		end;
	end;

procedure BoardAttack(attackerStatId: integer; x, y: integer);
	begin
		if (attackerStatId <> 0) then
			DamageStat(attackerStatId);

		if (attackerStatId > 0) and (attackerStatId <= CurrentStatTicked) then
			Dec(CurrentStatTicked);

		if (Board.Tiles[x][y].Element <> E_PLAYER) then begin
			BoardDamageTile(x, y);
			SoundQueue(2, SoundParse('4'));
		end;
	end;

procedure CalcDirectionRnd(var deltaX, deltaY: integer);
	begin
		deltaX := Random(3) - 1;

		if deltaX = 0 then
			deltaY := Random(2) * 2 - 1
		else
			deltaY := 0;
	end;

procedure CalcDirectionSeek(x, y: integer; var deltaX, deltaY: integer);
	begin
		deltaX := 0;
		deltaY := 0;

		if (Random(2) < 1) or (Board.Stats[0].Y = y) then
			deltaX := Signum(Board.Stats[0].X - x);

		if deltaX = 0 then
			deltaY := Signum(Board.Stats[0].Y - y);

	end;

procedure TransitionDrawBoardChange;
	begin
		TransitionDrawToFill(#219, $05);
		TransitionDrawToBoard;
	end;

procedure BoardEnter;
	begin
		Board.Info.StartPlayerX := Board.Stats[0].X;
		Board.Info.StartPlayerY := Board.Stats[0].Y;
		if OopSend(0,'ALL:BOARDENTER', false) then begin end;
		JustEntered := true;
	end;

procedure DrawKeyContents;
	var
		iy: integer;
	begin
		iy := 3;
		if WorldGetFlagPosition('BINBKEY') >= 0 then begin
			VideoWriteText(62, iy, $1D, 'Chambinbs Key  ');
			Inc(iy);
		end;
		if WorldGetFlagPosition('UPKEY') >= 0 then begin
			VideoWriteText(62, iy, $1D, 'Upstairs Key   ');
			Inc(iy);
		end;
		if WorldGetFlagPosition('HERMKEY') >= 0 then begin
			VideoWriteText(62, iy, $1D, 'Hermit Key     ');
			Inc(iy);
		end;
		if WorldGetFlagPosition('WHENKEY') >= 0 then begin
			VideoWriteText(62, iy, $1D, 'When Cellar Key');
			Inc(iy);
		end;
		if WorldGetFlagPosition('DINEKEY') >= 0 then begin
			VideoWriteText(62, iy, $1D, 'Dining Hall Key');
			Inc(iy);
		end;
		if WorldGetFlagPosition('PRIVKEY') >= 0 then begin
			VideoWriteText(62, iy, $1D, 'Private Key     ');
			Inc(iy);
		end;
	end;

procedure WipeKeyContents;
	var
		iy: integer;
	begin
		iy := 3;
		if WorldGetFlagPosition('BINBKEY') >= 0 then begin
			VideoWriteText(62, iy, $1D, '               ');
			Inc(iy);
		end;
		if WorldGetFlagPosition('UPKEY') >= 0 then begin
			VideoWriteText(62, iy, $1D, '               ');
			Inc(iy);
		end;
		if WorldGetFlagPosition('HERMKEY') >= 0 then begin
			VideoWriteText(62, iy, $1D, '               ');
			Inc(iy);
		end;
		if WorldGetFlagPosition('WHENKEY') >= 0 then begin
			VideoWriteText(62, iy, $1D, '               ');
			Inc(iy);
		end;
		if WorldGetFlagPosition('DINEKEY') >= 0 then begin
			VideoWriteText(62, iy, $1D, '               ');
			Inc(iy);
		end;
		if WorldGetFlagPosition('PRIVKEY') >= 0 then begin
			VideoWriteText(62, iy, $1D, '               ');
			Inc(iy);
		end;
		GameUpdateSidebar;
	end;

procedure DrawInventoryContents;
	var
		i, ix, iy: integer;
	begin
		ix := 0;
		iy := 0;
		{ draw cancel }
		VideoWriteText(3, 2, $14,#29#240#227#31);
		VideoWriteText(4, 3, $41,#233#234);
		VideoWriteText(3, 4, $14,#18#248#232#12);
		VideoWriteText(2, 5, $1C,'Cancel');
		{ draw items }
		for i := 1 to ITEM_MAX do begin
			if WorldGetFlagPosition(ItemNames[i]) >= 0 then begin
				ix := 1 - ix;
				if ix = 0 then
					Inc(iy);
				if iy < 4 then begin
					if (i = 9) then begin
						if World.Info.Torches > 0 then begin
							DrawInventoryItem(ix,iy,'ICE');
						end else begin
							if World.Info.Ammo > 0 then begin
								DrawInventoryItem(ix,iy,'HOTWATER');
							end else begin
								DrawInventoryItem(ix,iy,'WATER');
							end;
						end;
					end else begin
						DrawInventoryItem(ix,iy,ItemNames[i]);
					end;
				end;
			end;
		end;
	end;

procedure UseInventoryItem(x,y: integer);
	var
		i, ix, iy: integer;
	begin
		ix := 0;
		iy := 0;
		for i := 1 to ITEM_MAX do begin
			if WorldGetFlagPosition(ItemNames[i]) >= 0 then begin
				ix := 1 - ix;
				if ix = 0 then
					Inc(iy);
				if (x = ix) and (y = iy) then begin
					if OopSend(0,'ALL:USEANY', false) then begin end;
					if OopSend(0,'ALL:USE' + ItemClass[i], false) then begin end;
					if OopSend(0,'ALL:USE' + ItemNames[i], false) then begin end;
				end;
			end;
		end;
	end;

procedure DrawInventorySelect(x,y: integer);
	var
		ix, iy: integer;
	begin
		for ix := 0 to 1 do begin
			for iy := 0 to 3 do begin
				if (x = ix) and (y = iy) then begin
					VideoWriteText((ix * 10), (iy * 6) + 1, $1F,#218);
					VideoWriteText((ix * 10) + 9, (iy * 6) + 1, $1F,#191);
					VideoWriteText((ix * 10), (iy * 6) + 6, $1F,#192);
					VideoWriteText((ix * 10) + 9, (iy * 6) + 6, $1F,#217);
				end else begin
					VideoWriteText((ix * 10), (iy * 6) + 1, $1F,#32);
					VideoWriteText((ix * 10) + 9, (iy * 6) + 1, $1F,#32);
					VideoWriteText((ix * 10), (iy * 6) + 6, $1F,#32);
					VideoWriteText((ix * 10) + 9, (iy * 6) + 6, $1F,#32);
				end;
			end;
		end;
	end;

procedure GameInventoryPrompt;
	var
		selecting: integer;
		ix, iy: integer;
	begin
		for ix := 0 to 19 do begin
			for iy := 1 to (BOARD_HEIGHT - 1) do
				VideoWriteText(ix, iy, $1F,#32);
			TrueDelay(6);
		end;
			for iy := 0 to (BOARD_HEIGHT - 1) do
			VideoWriteText(20, iy, $1F,#179);
		VideoWriteText(0, 0, $1F,#24#25#26#27 + ':MOVE  ENTER:USE');
		ix := 0;
		iy := 0;
		selecting := 3;

		DrawKeyContents;
		DrawInventoryContents;		
		SlowInput;
		while (selecting > 0) do begin
			DrawInventorySelect(ix,iy);
			InputReadWaitKey;
			case InputKeyPressed of
				KEY_UP: Dec(iy);
				KEY_DOWN: Inc(iy);
				KEY_LEFT, KEY_RIGHT: ix := 1 - ix;
				KEY_ENTER: begin
					UseInventoryItem(ix,iy);
					selecting := 0;
				end;
				'i','I': begin
					TrueDelay(200);
					Dec(selecting);
				end;
				else selecting := 0 end;
			if iy < 0 then
				iy := 3;
			if iy > 3 then
				iy := 0;
		end;
		
		{BoardDrawBorder;}
		for ix := 0 to 21 do begin
			for iy := 1 to BOARD_HEIGHT do
				BoardDrawTile(22 - ix, iy);
			TrueDelay(6);
		end;
		WipeKeyContents;
		{TransitionDrawToBoard;}
		{FastInput;}
	end;

procedure GameDebugPrompt;
	var
		input: TString50;
		i: integer;
		toggle: boolean;
	begin
		input := '';
		SidebarClearLine(4);
		SidebarClearLine(5);

		PromptString(63, 5, $1E, $0F, 11, PROMPT_ANY, input);
		for i := 1 to Length(input) do
			input[i] := UpCase(input[i]);

		toggle := true;
		if (input[1] = '+') or (input[1] = '-') then begin
			if input[1] = '-' then
				toggle := false;
			input := Copy(input, 2, Length(input) - 1);

			if toggle = true then begin
				WorldSetFlag(input);
			end else
				WorldClearFlag(input);
		end;

		DebugEnabled := WorldGetFlagPosition('DEBUG') >= 0;

		World.Info.Register[2] := 10000;
		if input = 'HEALTH' then
			World.Info.Health := 100
		else if input = 'AMMO' then
			World.Info.Ammo := World.Info.Ammo + 500
		else if input = 'TORCHES' then
			World.Info.Torches := World.Info.Torches + 500
		else if input = 'GEMS' then
			World.Info.Gems := World.Info.Gems + 50
		else if input = 'FLAGS' then
			WorldGetFlagList
		else if input = 'FREE' then
			Board.Stats[0].P2 := 0
		else if input = 'DARK' then begin
			Board.Info.IsDark := 1;
			TransitionDrawToBoard;
		end else if input = 'DIM' then begin
			Board.Info.IsDark := 3;
			TransitionDrawToBoard;
		end else if input = 'DANK' then begin
			Board.Info.IsDark := 4;
			TransitionDrawToBoard;
		end else if input = 'DEEP' then begin
			Board.Info.IsDark := 5;
			TransitionDrawToBoard;
		end else if input = 'LIGHT' then begin
			Board.Info.IsDark := 0;
			TransitionDrawToBoard;
		end else if input = 'BRIGHT' then begin
			Board.Info.IsDark := 2;
			TransitionDrawToBoard;			
		end else if input = 'ZAP' then begin
			for i := 0 to 3 do begin
				BoardDamageTile(Board.Stats[0].X + NeighborDeltaX[i], Board.Stats[0].Y + NeighborDeltaY[i]);
				Board.Tiles[Board.Stats[0].X + NeighborDeltaX[i]][Board.Stats[0].Y + NeighborDeltaY[i]].Element := E_EMPTY;
				BoardDrawTile(Board.Stats[0].X + NeighborDeltaX[i], Board.Stats[0].Y + NeighborDeltaY[i]);
			end;
		end;

		SoundQueue(10, #39#1);
		SidebarClearLine(4);
		SidebarClearLine(5);
		GameUpdateSidebar;
	end;

procedure GameAboutScreen;
	begin
		TextWindowDisplayFile('ABOUT.HLP', 'MIT License');
	end;

procedure GameDrawSidebar;
	begin
		SidebarClear;
		VideoWriteText(61, 0, $1F, '    - - - - -      ');
		if WorldGetFlagPosition('ADVOCAT') >= 0 then
			VideoWriteText(61, 1, $70, ' P L A S T I N B ')
		else
			VideoWriteText(62, 1, $70, ' WeaveZZT 1.4t ');
		VideoWriteText(61, 2, $1F, '    - - - - -      ');
		if GameStateElement = E_PLAYER then begin
			VideoWriteText(64, 9, $1E, ' Health:');
			{ VideoWriteText(64, 8, $1E, '   Ammo:');
			VideoWriteText(64, 9, $1E, 'Torches:'); }
			VideoWriteText(63, 10, $1E, 'Stickers:');
			VideoWriteText(64, 11, $1E, ' Points:');
			VideoWriteText(62, 12, $1E, 'Bag Space:');
			{ VideoWriteText(62, 9, $1F, ElementDefs[E_PLAYER].Character);
			VideoWriteText(62, 8, $1B, ElementDefs[E_AMMO].Character);
			VideoWriteText(62, 9, $16, ElementDefs[E_TORCH].Character);
			VideoWriteText(62, 10, $1B, ElementDefs[E_GEM].Character);
			VideoWriteText(62, 12, $1F, ElementDefs[E_KEY].Character); }
			VideoWriteText(62, 16, $70, ' I ');
			VideoWriteText(65, 16, $1F, ' Inventory');
			VideoWriteText(62, 17, $30, ' B ');
			VideoWriteText(62, 21, $70, ' H ');
			VideoWriteText(65, 21, $1F, ' Hint');
			VideoWriteText(67, 18, $30, ' '#24#25#26#27' ');
			VideoWriteText(72, 18, $1F, ' Move');
			VideoWriteText(62, 22, $30, ' P ');
			VideoWriteText(65, 22, $1F, ' Pause');
			VideoWriteText(62, 23, $70, ' Q ');
			VideoWriteText(65, 23, $1F, ' Quit');
		end else if GameStateElement = E_MONITOR then begin
			SidebarPromptSlider(false, 66, 21, 'Game speed:;FS', TickSpeed);
			VideoWriteText(62, 21, $70, ' S ');
			VideoWriteText(62, 7, $30, ' W ');
			VideoWriteText(65, 7, $1E, ' World:');

			if Length(World.Info.Name) <> 0 then
				VideoWriteText(69, 8, $1F, World.Info.Name)
			else
				VideoWriteText(69, 8, $1F, 'Untitled');

		if SaveTest then begin
			VideoWriteText(62, 11, $70, ' P ');
			VideoWriteText(65, 11, $1E, ' Play');
		end;
			VideoWriteText(62, 12, $30, ' N ');
			VideoWriteText(65, 12, $1F, ' New Game');
			VideoWriteText(62, 13, $70, ' Q ');
			VideoWriteText(65, 13, $1E, ' Quit');
			VideoWriteText(62, 16, $30, ' A ');
			VideoWriteText(65, 16, $1F, ' MIT License');
			VideoWriteText(62, 17, $70, ' H ');
			VideoWriteText(65, 17, $1E, ' Hall of Fame');

		end;
	end;

procedure GamePlayLoop(boardChanged: boolean);
	var
		exitLoop: boolean;
		pauseBlink: boolean;
	begin
		GameDrawSidebar;
		GameUpdateSidebar;

		if JustStarted then begin
			if Length(StartupWorldFileName) <> 0 then begin
				SidebarClearLine(8);
				VideoWriteText(69, 8, $1F, StartupWorldFileName);
				if not WorldLoad('BINB', '.ZZT') then WorldCreate;
			end;
			ReturnBoardId := World.Info.CurrentBoard;
			BoardChange(0);
			JustStarted := false;
		end;

		Board.Tiles[Board.Stats[0].X][Board.Stats[0].Y].Element := GameStateElement;
		Board.Tiles[Board.Stats[0].X][Board.Stats[0].Y].Color := ElementDefs[GameStateElement].Color;

		if GameStateElement = E_MONITOR then begin
			DisplayMessage(0, '');
			VideoWriteText(62, 5, $1B, 'Pick a command:');
		end;

		if boardChanged then begin
			TransitionDrawBoardChange;
			GameFadeIn(Board.Info.IsDark);
		end;

		TickTimeDuration := TickSpeed * 2;
		GamePlayExitRequested := false;
		exitLoop := false;

		CurrentTick := Random(100);
		CurrentStatTicked := Board.StatCount + 1;

		repeat
			if GamePaused then begin
				{GameUpdateSidebar;}
				if SoundHasTimeElapsed(TickTimeCounter, 25) then
					pauseBlink := not pauseBlink;

				if pauseBlink then begin
					if ConfigPaper then begin
						if World.Info.PlayerColor = $09 then begin
							VideoWriteText(Board.Stats[0].X - 1, Board.Stats[0].Y - 1,
								$F9, ElementDefs[E_PLAYER].Character);
						end else begin
							VideoWriteText(Board.Stats[0].X - 1, Board.Stats[0].Y - 1,
								World.Info.PlayerColor, ElementDefs[E_PLAYER].Character);
						end;
					end else begin
						if World.Info.PlayerColor = $00 then
							World.Info.PlayerColor := $1F;
						VideoWriteText(Board.Stats[0].X - 1, Board.Stats[0].Y - 1,
							World.Info.PlayerColor, ElementDefs[E_PLAYER].Character);
					end;
				end else begin
					if Board.Tiles[Board.Stats[0].X][Board.Stats[0].Y].Element = E_PLAYER then begin
						if ConfigPaper then begin
							VideoWriteText(Board.Stats[0].X - 1, Board.Stats[0].Y - 1, $F0, ' ');
						end else begin
							VideoWriteText(Board.Stats[0].X - 1, Board.Stats[0].Y - 1, $0F, ' ');
						end;
					end else begin
						BoardDrawTile(Board.Stats[0].X, Board.Stats[0].Y);
					end;
				end;

				VideoWriteText(64, 5, $1F, 'Pausing...');
				InputUpdate;

				if InputCheck(#27) then
					GamePromptEndPlay;
				if (InputDeltaX <> 0) or (InputDeltaY <> 0) then begin
					{ Unpause }
					GamePaused := false;
					SidebarClearLine(5);
					CurrentTick := Random(100);
					CurrentStatTicked := Board.StatCount + 1;
				end;

			end else begin { not GamePaused }
				if CurrentStatTicked <= Board.StatCount then begin
					with Board.Stats[CurrentStatTicked] do begin
						if (Cycle <> 0) and ((CurrentTick mod Cycle) = (CurrentStatTicked mod Cycle)) then
							ElementDefs[Board.Tiles[X][Y].Element].TickProc(CurrentStatTicked);

						CurrentStatTicked := CurrentStatTicked + 1;
					end;
				end;
			end;

			if (CurrentStatTicked > Board.StatCount) and not GamePlayExitRequested then begin
				{ all stats ticked }
				if SoundHasTimeElapsed(TickTimeCounter, TickTimeDuration) then begin
					{ next cycle }
					CurrentTick := CurrentTick + 1;
					if CurrentTick > 420 then
						CurrentTick := 1;
					CurrentStatTicked := 0;
					if (Board.Info.ReenterWhenZapped = false) then
						FastInput;
					InputUpdate;
				end;
			end;
		until (exitLoop or GamePlayExitRequested) and GamePlayExitRequested;

		SoundClearQueue;

		if GameStateElement = E_PLAYER then begin
			if World.Info.Health <= 0 then begin
				HighScoresAdd(World.Info.Score);
			end;
		end else if GameStateElement = E_MONITOR then begin
			SidebarClearLine(5);
		end;

		Board.Tiles[Board.Stats[0].X][Board.Stats[0].Y].Element := E_PLAYER;
		Board.Tiles[Board.Stats[0].X][Board.Stats[0].Y].Color := World.Info.PlayerColor;

		SoundBlockQueueing := false;
	end;

procedure GameTitleLoop;
	var
		boardChanged: boolean;
		f: file;
		i: integer;
		fn: TString50;
	begin
		GameTitleExitRequested := false;
		JustStarted := true;
		ReturnBoardId := 0;
		boardChanged := true;
		repeat
			BoardChange(0);
			repeat
				GameStateElement := E_MONITOR;
				GamePaused := false;
				GamePlayLoop(boardChanged);
				boardChanged := false;

				case UpCase(InputKeyPressed) of
					'N': begin
						if SidebarPromptYesNo('Start New Game? ', false) and (InputKeyPressed <> KEY_ESCAPE) then begin
							for i := 1 to 9 do begin
								Str(i, fn);
								if WorldLoad('BINB' + fn, '.ZZT') then begin
									WorldSave(i);
								end
							end;
							if WorldLoad('BINB', '.ZZT') then begin
								BoardChange(World.Info.CurrentBoard);
								BoardEnter;
								GameStateElement := E_PLAYER;
								GamePaused := true;
								GamePlayLoop(true);
								boardChanged := true;
							end;
						end;
					end;
					'P': begin
						if WorldLoad('BINBSAVE', '.S99') then begin 
							GameFadeOut(1);
							BoardChange(World.Info.CurrentBoard);
							BoardEnter;
							GameStateElement := E_PLAYER;
							GamePaused := true;
							if (World.Info.PlayerAppearance > 0) and (World.Info.PlayerAppearance < 3) then
								SetExpr(World.Info.PlayerAppearance, 0);
							If World.Info.Health = 0 then
								Inc(World.Info.Health);
							GamePlayLoop(true);
							boardChanged := true;
						end;
					end;
					'A': begin
						GameAboutScreen;
					end;
					'S': begin
						SidebarPromptSlider(true, 66, 21, 'Game speed:;FS', TickSpeed);
						InputKeyPressed := #0;
					end;
					'H': begin
						HighScoresLoad;
						HighScoresDisplay(1);
					end;
					{'|': begin
						GameDebugPrompt;
					end; }
					KEY_ESCAPE, 'Q': begin
						GameTitleExitRequested := SidebarPromptYesNo('Quit ZZT? ', true);
					end;
				end;

			until boardChanged or GameTitleExitRequested;
		until GameTitleExitRequested;
	end;

begin
end.
